@@ -44,6 +44,7 @@ struct _HeMenuViewPrivate
 
 	GtkWidget *menu_widget;
 	gboolean do_activate;
+	GQueue *parents_q;
 };
 
 enum {
@@ -56,6 +57,7 @@ G_DEFINE_TYPE(HeMenuView, he_menu_view, GTK_TYPE_FIXED);
 
 static void hmvc_notify_menu_widget(GtkWidget *hmvc, GParamSpec *pspec, HeMenuView *hmv);
 static gboolean my_toggle_menu(HildonWindow *self, guint button, guint32 time);
+static void hmvc_destroy_cb(GtkWidget *hmvc, HeMenuView *hmv);
 
 /*
  * Start resizing the widget
@@ -75,6 +77,26 @@ start_resizing(HeMenuView *hmv)
 	}
 }
 
+static void
+redo_parents_q(HeMenuView *hmv)
+{
+	GtkWidget *parent;
+	GList *ll;
+
+	for (ll = hmv->priv->parents_q->head ; ll ; ll = ll->next)
+		if (ll->data)
+			g_object_remove_weak_pointer(G_OBJECT(ll->data), &(ll->data));
+	g_queue_clear(hmv->priv->parents_q);
+
+	if (hmv->priv->menu_q->head)
+		for (parent = _hmv_private_menu_get_parent_menu(he_menu_view_column_get_menu(HE_MENU_VIEW_COLUMN(hmv->priv->menu_q->head->data))) ;
+		     parent ;
+		     parent = _hmv_private_menu_get_parent_menu(parent)) {
+			g_queue_push_tail(hmv->priv->parents_q, parent);
+			g_object_add_weak_pointer(G_OBJECT(hmv->priv->parents_q->tail->data), (gpointer *)&(hmv->priv->parents_q->tail->data));
+		}
+}
+
 /*
  * Finish resizing the widget
  */
@@ -87,10 +109,11 @@ stop_resizing(HeMenuView *hmv)
 		hmv->priv->resizing = FALSE ;
 		if (hmv->priv->froze_updates && wnd) {
 			gdk_window_thaw_updates(wnd);
-//			g_print("size_allocate: +++++ enable updates +++++\n");
 			hmv->priv->froze_updates = FALSE;
 		}
 	}
+
+	redo_parents_q(hmv);
 }
 
 /*
@@ -107,9 +130,59 @@ add_new_column(HeMenuView *hmv, GtkMenu *menu)
 	gtk_container_add(GTK_CONTAINER(hmv), hmvc);
 	g_signal_connect(G_OBJECT(hmvc), "notify::menu-widget", (GCallback)hmvc_notify_menu_widget, hmv);
 
+	g_object_set_data(G_OBJECT(hmvc), "destroy-signal-id",
+		GINT_TO_POINTER(g_signal_connect(G_OBJECT(hmvc), "destroy", (GCallback)hmvc_destroy_cb, hmv)));
+
 	return hmvc;
 }
 
+/*
+ * Move up menu levels
+ */
+static void
+button_clicked(GtkWidget *button, HeMenuView *hmv)
+{
+	GList *ll;
+
+	for (ll = hmv->priv->parents_q->head ; ll ; ll = ll->next)
+		if (ll->data)
+			break;
+
+	if (ll) {
+		g_queue_push_head(hmv->priv->menu_q, add_new_column(hmv, GTK_MENU(ll->data)));
+		hmv->priv->head_is_new = TRUE;
+		redo_parents_q(hmv);
+	}
+
+	if (hmv->priv->parents_q->length <= 0)
+		gtk_widget_hide(hmv->priv->button);
+
+	if (hmv->priv->menu_q->length <= 0)
+		gtk_widget_destroy(GTK_WIDGET(hmv));
+}
+
+/*
+ * Clean up column queue when column is destroyed
+ */
+static void
+hmvc_destroy_cb(GtkWidget *hmvc, HeMenuView *hmv)
+{
+	g_queue_remove(hmv->priv->menu_q, hmvc);
+}
+
+/*
+ * Destroy a column
+ */
+static void
+destroy_hmvc(GObject *hmvc)
+{
+	guint destroy_signal_id = GPOINTER_TO_INT(g_object_get_data(hmvc, "destroy-signal-id"));
+
+	if (destroy_signal_id)
+		g_signal_handler_disconnect(hmvc, destroy_signal_id);
+	gtk_widget_destroy(GTK_WIDGET(hmvc));
+}
+
 /*
  * Handle notifications about clicks on menu items
  */
@@ -125,7 +198,7 @@ hmvc_notify_menu_widget(GtkWidget *hmvc, GParamSpec *pspec, HeMenuView *hmv)
 
 		if (submenu) {
 			while(hmv->priv->menu_q->tail->data != hmvc) {
-				gtk_widget_destroy(GTK_WIDGET(hmv->priv->menu_q->tail->data));
+				destroy_hmvc(G_OBJECT(hmv->priv->menu_q->tail->data));
 				g_queue_pop_tail(hmv->priv->menu_q);
 			}
 
@@ -139,27 +212,6 @@ hmvc_notify_menu_widget(GtkWidget *hmvc, GParamSpec *pspec, HeMenuView *hmv)
 	}
 }
 
-/*
- * Move up menu levels
- */
-static void
-button_clicked(GtkWidget *button, HeMenuView *hmv)
-{
-	if (hmv->priv->menu_q->head) {
-		GtkWidget *widget = _hmv_private_menu_get_parent_menu(he_menu_view_column_get_menu(HE_MENU_VIEW_COLUMN(hmv->priv->menu_q->head->data)));
-
-		if (widget) {
-			g_queue_push_head(hmv->priv->menu_q, add_new_column(hmv, GTK_MENU(widget)));
-			hmv->priv->head_is_new = TRUE;
-
-			if (!_hmv_private_menu_get_parent_menu(he_menu_view_column_get_menu(HE_MENU_VIEW_COLUMN(hmv->priv->menu_q->head->data))) && GTK_WIDGET_VISIBLE(hmv->priv->button)) {
-				gtk_widget_hide(hmv->priv->button);
-				return;
-			}
-		}
-	}
-}
-
 /*
  * Allocate the space for a single column
  */
@@ -226,7 +278,7 @@ destroy_excess_columns(HeMenuView *hmv, int alloc_width, gboolean tail_is_signif
 			/* Never empty the queue completely */
 			if (hmv->priv->menu_q->length > 1) {
 				keep_looping = (ll_itr != (*p_end));
-				gtk_widget_destroy(GTK_WIDGET((*p_end)->data));
+				destroy_hmvc(G_OBJECT((*p_end)->data));
 				(*pop_fn)(hmv->priv->menu_q);
 			}
 			else
@@ -290,8 +342,12 @@ size_allocate(GtkWidget *widget, GtkAllocation *alloc)
 	}
 
 	if (!is_dirty) {
-		stop_resizing(hmv);
-		hmv->priv->head_is_new = TRUE;
+		if (0 == hmv->priv->menu_q->length)
+			button_clicked(hmv->priv->button, hmv);
+		else {
+			stop_resizing(hmv);
+			hmv->priv->head_is_new = TRUE;
+		}
 	}
 }
 
@@ -394,6 +450,7 @@ he_menu_view_init(HeMenuView *hmv)
 	hmv->priv->resizing      = FALSE;
 	hmv->priv->froze_updates = FALSE;
 	hmv->priv->menu_q        = g_queue_new();
+	hmv->priv->parents_q     = g_queue_new();
 	hmv->priv->button        =
 		g_object_new(GTK_TYPE_BUTTON, "image",
 			g_object_new(GTK_TYPE_IMAGE, "visible", TRUE, "stock", GTK_STOCK_GO_BACK, "icon-size", GTK_ICON_SIZE_MENU, NULL),
@@ -424,7 +481,7 @@ he_menu_view_set_menu(HeMenuView *hmv, GtkMenu *menu)
 			hmv->priv->menu = NULL;
 		}
 
-		g_queue_foreach(hmv->priv->menu_q, (GFunc)gtk_widget_destroy, NULL);
+		g_queue_foreach(hmv->priv->menu_q, (GFunc)destroy_hmvc, NULL);
 		g_queue_clear(hmv->priv->menu_q);
 
 		if (menu) {
@@ -499,6 +556,7 @@ typedef struct {
 	GtkWidget *dlg;
 	GtkMenuItem *item;
 	GtkMenu *menu;
+	GtkWidget *hmv;
 } DialogHasDiedParams;
 
 /*
@@ -532,34 +590,41 @@ menu_widget_notify(HeMenuView *hmv, GParamSpec *pspec, DialogHasDiedParams *para
 	gtk_widget_destroy(params->dlg);
 }
 
+static void
+hmv_destroy(GtkWidget *hmv, DialogHasDiedParams *params)
+{
+	params->menu = NULL;
+	gtk_widget_destroy(params->dlg);
+}
+
 /*
  * Pop a dialog displaying the given #GtkMenu and return if @run is FALSE.
  */
 static void
 pop_dialog(GtkWindow *parent, GtkMenu *menu, gboolean run)
 {
-	GtkWidget *hmv;
 	GtkWidget *dlg;
 	DialogHasDiedParams *params = g_new0(DialogHasDiedParams, 1);
 
-	hmv = g_object_new(HE_TYPE_MENU_VIEW, "visible", TRUE, "menu", menu, "do-activate", FALSE, NULL);
+	params->hmv = g_object_new(HE_TYPE_MENU_VIEW, "visible", TRUE, "menu", menu, "do-activate", FALSE, NULL);
 
 	dlg = g_object_new(GTK_TYPE_DIALOG, "visible", TRUE, "transient-for", parent, NULL);
 	g_object_set_data_full(G_OBJECT(dlg), "no-close-timeout-id", 
 		GINT_TO_POINTER(g_timeout_add_seconds(1, (GSourceFunc)dialog_ignore_close, dlg)),
 		(GDestroyNotify)g_source_remove);
 	g_signal_connect(G_OBJECT(dlg), "delete-event", (GCallback)dlg_ignore_delete, NULL);
+	g_signal_connect(G_OBJECT(params->hmv), "destroy", (GCallback)hmv_destroy, params);
 
 	params->dlg = dlg;
 	params->menu = g_object_ref(menu);
 
 	unset_borders(GTK_BIN(dlg)->child, NULL);
 
-	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), hmv);
+	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), params->hmv);
 
 	g_object_weak_ref(G_OBJECT(dlg), (GWeakNotify)dialog_has_died, params);
 
-	g_signal_connect(G_OBJECT(hmv), "notify::menu-widget", (GCallback)menu_widget_notify, params);
+	g_signal_connect(G_OBJECT(params->hmv), "notify::menu-widget", (GCallback)menu_widget_notify, params);
 
 	if (run)
 		gtk_dialog_run(GTK_DIALOG(dlg));
